"""Core constructs for Jac Language."""

from __future__ import annotations

from collections.abc import Callable
from dataclasses import asdict, dataclass, field, fields, is_dataclass
from enum import IntEnum
from functools import cached_property
from inspect import _empty, signature
from logging import getLogger
from pickle import dumps
from types import UnionType
from typing import Any, ClassVar, TypeAlias, TypeVar
from uuid import UUID, uuid4

from jaclang.pycore.constant import EdgeDir

logger = getLogger(__name__)
TARCH = TypeVar("TARCH", bound="Archetype")
TANCH = TypeVar("TANCH", bound="Anchor")
T = TypeVar("T")


class AccessLevel(IntEnum):
    """Access level enum."""

    NO_ACCESS = -1
    READ = 0
    CONNECT = 1
    WRITE = 2

    @staticmethod
    def cast(val: int | str | AccessLevel) -> AccessLevel:
        """Cast access level."""
        if isinstance(val, int):
            return AccessLevel(val)
        elif isinstance(val, str):
            return AccessLevel[val.upper()]
        else:
            return val


@dataclass
class Access:
    """Access Structure."""

    anchors: dict[str, AccessLevel] = field(default_factory=dict)

    def check(self, anchor: str) -> AccessLevel | None:
        """Validate access."""
        return self.anchors.get(anchor)


@dataclass
class Permission:
    """Anchor Access Handler."""

    all: AccessLevel = AccessLevel.NO_ACCESS
    roots: Access = field(default_factory=Access)


@dataclass
class AnchorReport:
    """Report Handler."""

    id: str
    context: dict[str, Any]


ObjectSpatialFilter: TypeAlias = (
    "Callable[[Archetype], bool] | Archetype | list[Archetype] | None"
)


@dataclass(eq=False, repr=False)
class ObjectSpatialDestination:
    """Object-Spatial Destination."""

    direction: EdgeDir
    edge: Callable[[Archetype], bool] | None = None
    node: Callable[[Archetype], bool] | None = None

    def edge_filter(self, arch: Archetype) -> bool:
        """Filter edge."""
        return not self.edge or self.edge(arch)

    def node_filter(self, arch: Archetype) -> bool:
        """Filter node."""
        return not self.node or self.node(arch)


@dataclass(eq=False, repr=False)
class ObjectSpatialPath:
    """Object-Spatial Path."""

    origin: list[NodeArchetype]
    destinations: list[ObjectSpatialDestination]
    edge_only: bool
    from_visit: bool

    def __init__(
        self,
        origin: NodeArchetype | list[NodeArchetype],
        destinations: list[ObjectSpatialDestination] | None = None,
    ) -> None:
        """Override Init."""
        if not isinstance(origin, list):
            origin = [origin]
        self.origin = origin
        self.destinations = [] if destinations is None else destinations
        self.edge_only = False
        self.from_visit = False

    def convert(
        self, filter: ObjectSpatialFilter
    ) -> Callable[[Archetype], bool] | None:
        """Convert filter."""
        if not filter:
            return None
        if callable(filter):
            return filter
        elif isinstance(filter, list):
            return lambda i: i in filter
        return lambda i: i == filter

    def append(
        self,
        direction: EdgeDir,
        edge: ObjectSpatialFilter,
        node: ObjectSpatialFilter,
    ) -> ObjectSpatialPath:
        """Append destination."""
        self.destinations.append(
            ObjectSpatialDestination(direction, self.convert(edge), self.convert(node))
        )
        return self

    def edge_out(
        self,
        edge: ObjectSpatialFilter = None,
        node: ObjectSpatialFilter = None,
    ) -> ObjectSpatialPath:
        """Override greater than function."""
        return self.append(EdgeDir.OUT, edge, node)

    def edge_in(
        self,
        edge: ObjectSpatialFilter = None,
        node: ObjectSpatialFilter = None,
    ) -> ObjectSpatialPath:
        """Override greater than function."""
        return self.append(EdgeDir.IN, edge, node)

    def edge_any(
        self,
        edge: ObjectSpatialFilter = None,
        node: ObjectSpatialFilter = None,
    ) -> ObjectSpatialPath:
        """Override greater than function."""
        return self.append(EdgeDir.ANY, edge, node)

    def set_edge_only(self) -> ObjectSpatialPath:
        """Set edge only."""
        self.edge_only = True
        return self

    def edge(self) -> ObjectSpatialPath:
        """Set edge only mode (alias for set_edge_only)."""
        self.edge_only = True
        return self

    def visit(self) -> ObjectSpatialPath:
        """Set from visit mode."""
        self.from_visit = True
        return self

    # Aliases for backwards compatibility
    set_visit = visit

    def repr_builder(
        self, repr_str: str, dest: ObjectSpatialDestination, mark: str
    ) -> str:
        """Repr builder."""
        repr_str += mark
        repr_str += f" (edge{' filter' if dest.edge else ''}) "
        repr_str += mark
        repr_str += f" (node{' filter' if dest.node else ''}) "
        return repr_str

    def __repr__(self) -> str:
        """Override repr."""
        repr_str = "nodes "
        for dest in self.destinations:
            if dest.direction == EdgeDir.IN:
                repr_str = self.repr_builder(repr_str, dest, "<<")
            elif dest.direction == EdgeDir.OUT:
                repr_str = self.repr_builder(repr_str, dest, ">>")
            else:
                repr_str = self.repr_builder(repr_str, dest, "--")
        return repr_str.strip()


@dataclass(eq=False, repr=False, kw_only=True)
class Anchor:
    """Object Anchor."""

    archetype: Archetype
    id: UUID = field(default_factory=uuid4)
    root: UUID | None = None
    access: Permission = field(default_factory=Permission)
    persistent: bool = False
    hash: int = 0

    def is_populated(self) -> bool:
        """Check if state."""
        return "archetype" in self.__dict__

    def make_stub(self: TANCH) -> TANCH:
        """Return unsynced copy of anchor."""
        if self.is_populated():
            unloaded = object.__new__(self.__class__)
            unloaded.id = self.id
            return unloaded
        return self

    def populate(self) -> None:
        """Retrieve the Archetype from db and return."""
        from jaclang import JacRuntimeInterface as Jac

        ctx = Jac.get_context()
        # Orchestrator handles read-through caching (L1 -> L3)
        anchor = ctx.mem.get(self.id)
        if anchor:
            self.__dict__.update(anchor.__dict__)

    def __getattr__(self, name: str) -> object:
        """Trigger load if detects unloaded state."""
        if not self.is_populated():
            self.populate()
            if not self.is_populated():
                raise ValueError(
                    f"{self.__class__.__name__} [{self.id}] is not a valid reference!"
                )
            return getattr(self, name)
        raise AttributeError(
            f"'{self.__class__.__name__} object has not attribute {name}'"
        )

    def __getstate__(self) -> dict[str, Any]:
        """Serialize Anchor."""
        if self.is_populated():
            unlinked = object.__new__(self.archetype.__class__)
            unlinked.__dict__.update(self.archetype.__dict__)
            unlinked.__dict__.pop("__jac__", None)
            return {
                "id": self.id,
                "archetype": unlinked,
                "root": self.root,
                "access": self.access,
                "persistent": self.persistent,
            }
        else:
            return {"id": self.id}

    def __setstate__(self, state: dict[str, Any]) -> None:
        """Deserialize Anchor."""
        self.__dict__.update(state)
        if self.is_populated() and self.archetype:
            self.archetype.__jac__ = self  # type: ignore[misc]
            self.hash = hash(dumps(self))

    def __repr__(self) -> str:
        """Override representation."""
        if self.is_populated():
            attrs = ""
            for f in fields(self):
                if f.name in self.__dict__:
                    attrs += f"{f.name}={self.__dict__[f.name]}, "
            attrs = attrs[:-2]
        else:
            attrs = f"id={self.id}"
        return f"{self.__class__.__name__}({attrs})"

    def report(self) -> AnchorReport:
        """Report Anchor."""
        return AnchorReport(
            id=self.id.hex,
            context=asdict(self.archetype)
            if is_dataclass(self.archetype) and not isinstance(self.archetype, type)
            else {},
        )

    def __hash__(self) -> int:
        """Override hash for anchor."""
        return hash(self.id)

    def __eq__(self, other: object) -> bool:
        """Override equal implementation."""
        if isinstance(other, Anchor):
            return self.__class__ is other.__class__ and self.id == other.id
        return False


@dataclass(eq=False, repr=False, kw_only=True)
class NodeAnchor(Anchor):
    """Node Anchor."""

    archetype: NodeArchetype  # type: ignore[assignment]
    edges: list[EdgeAnchor]

    def __getstate__(self) -> dict[str, object]:
        """Serialize Node Anchor."""
        state = super().__getstate__()
        if self.is_populated():
            state["edges"] = [edge.make_stub() for edge in self.edges]
        return state


@dataclass(eq=False, repr=False, kw_only=True)
class EdgeAnchor(Anchor):
    """Edge Anchor."""

    archetype: EdgeArchetype  # type: ignore[assignment]
    source: NodeAnchor
    target: NodeAnchor
    is_undirected: bool

    def __getstate__(self) -> dict[str, object]:
        """Serialize Edge Anchor."""
        state = super().__getstate__()
        if self.is_populated():
            state.update(
                {
                    "source": self.source.make_stub(),
                    "target": self.target.make_stub(),
                    "is_undirected": self.is_undirected,
                }
            )
        return state


@dataclass(eq=False, repr=False, kw_only=True)
class WalkerAnchor(Anchor):
    """Walker Anchor."""

    archetype: WalkerArchetype  # type: ignore[assignment]
    path: list[NodeAnchor | EdgeAnchor] = field(default_factory=list)
    next: list[NodeAnchor | EdgeAnchor] = field(default_factory=list)
    ignores: list[NodeAnchor] = field(default_factory=list)
    disengaged: bool = False


@dataclass(eq=False, repr=False, kw_only=True)
class ObjectAnchor(Anchor):
    """Object Anchor."""

    archetype: ObjectArchetype  # type: ignore[assignment]


@dataclass(eq=False, repr=False, kw_only=True)
class Archetype:
    """Archetype Protocol."""

    _jac_entry_funcs_: ClassVar[list[ObjectSpatialFunction]] = []
    _jac_exit_funcs_: ClassVar[list[ObjectSpatialFunction]] = []

    @cached_property
    def __jac__(self) -> Anchor:
        """Create default anchor."""
        return Anchor(archetype=self)

    def __init_subclass__(cls, **kwargs: object) -> None:
        """Configure subclasses."""
        super().__init_subclass__(**kwargs)
        if not cls.__dict__.get("__jac_base__", False):
            from jaclang import JacRuntimeInterface as _

            _.make_archetype(cls)

    def __repr__(self) -> str:
        """Override repr for archetype."""
        return f"{self.__class__.__name__}"

    def __jac_access__(self) -> AccessLevel | str | int | None:
        """Override access validation."""
        return None


class NodeArchetype(Archetype):
    """Node Archetype Protocol."""

    __jac_base__: ClassVar[bool] = True

    @cached_property
    def __jac__(self) -> NodeAnchor:
        """Create default anchor."""
        return NodeAnchor(archetype=self, edges=[])


class EdgeArchetype(Archetype):
    """Edge Archetype Protocol."""

    __jac_base__: ClassVar[bool] = True
    __jac__: EdgeAnchor


class WalkerArchetype(Archetype):
    """Walker Archetype Protocol."""

    __jac_async__: ClassVar[bool] = False
    __jac_base__: ClassVar[bool] = True
    reports: list[Any] = field(default_factory=list)

    @cached_property
    def __jac__(self) -> WalkerAnchor:
        """Create default anchor."""
        return WalkerAnchor(archetype=self)


class ObjectArchetype(Archetype):
    """Object Archetype Protocol."""

    __jac_base__: ClassVar[bool] = True

    @cached_property
    def __jac__(self) -> ObjectAnchor:
        """Create default anchor."""
        return ObjectAnchor(archetype=self)


@dataclass(eq=False)
class GenericEdge(EdgeArchetype):
    """Generic Edge."""

    __jac_base__: ClassVar[bool] = True

    def __repr__(self) -> str:
        """Override repr for archetype."""
        return f"{self.__class__.__name__}()"


@dataclass(eq=False)
class Root(NodeArchetype):
    """Generic Root Node."""

    __jac_base__: ClassVar[bool] = True

    @cached_property
    def __jac__(self) -> NodeAnchor:
        """Create default anchor."""
        return NodeAnchor(archetype=self, persistent=True, edges=[])

    def __repr__(self) -> str:
        """Override repr for archetype."""
        return f"{self.__class__.__name__}()"


@dataclass(eq=False)
class ObjectSpatialFunction:
    """Object-Spatial Function."""

    name: str
    func: Callable[[Any, Any], Any]

    @cached_property
    def trigger(self) -> type | UnionType | tuple[type | UnionType, ...] | None:
        """Get function parameter annotations."""
        parameters = signature(self.func, eval_str=True).parameters
        if len(parameters) >= 2:
            second_param = list(parameters.values())[1]
            ty = second_param.annotation
            return ty if ty != _empty else None
        return None



"""Jac Semantic Analysis Pass."""

import ast as ast3
from collections.abc import Sequence

import jaclang.pycore.unitree as uni
from jaclang.pycore.constant import Tokens as Tok
from jaclang.pycore.passes.uni_pass import UniPass


class SemanticAnalysisPass(UniPass):
    """Jac Semantic Analysis Pass."""

    def enter_archetype(self, node: uni.Archetype) -> None:
        def inform_from_walker(node: uni.UniNode) -> None:
            for i in (
                node.get_all_sub_nodes(uni.VisitStmt)
                + node.get_all_sub_nodes(uni.DisengageStmt)
                + node.get_all_sub_nodes(uni.EdgeOpRef)
                + node.get_all_sub_nodes(uni.EventSignature)
                + node.get_all_sub_nodes(uni.TypedCtxBlock)
            ):
                i.from_walker = True

        if node.arch_type.name == Tok.KW_WALKER:
            inform_from_walker(node)
            for i in self.get_all_sub_nodes(node, uni.Ability):
                if isinstance(i.body, uni.ImplDef):
                    inform_from_walker(i.body)

        # Validate archetype structure
        self._check_archetype(node)

    def _check_archetype(self, node: uni.Archetype) -> None:
        """Check a single archetype for issues with attributes and methods."""
        if node.arch_type.name != Tok.KW_CLASS and isinstance(node.body, Sequence):
            # Check for non-default attributes following default attributes
            found_default_init = False
            for stmnt in node.body:
                if not isinstance(stmnt, uni.ArchHas):
                    continue
                for var in stmnt.vars:
                    if (var.value is not None) or (var.defer):
                        found_default_init = True
                    else:
                        if found_default_init:
                            self.log_error(
                                f"Non default attribute '{var.name.value}' follows default attribute",
                                node_override=var.name,
                            )
                            break

            # Check for postinit requirement with deferred attributes
            post_init_vars: list[uni.HasVar] = []
            postinit_method: uni.Ability | None = None

            for item in node.body:
                if isinstance(item, uni.ArchHas):
                    for var in item.vars:
                        if var.defer:
                            post_init_vars.append(var)

                elif isinstance(item, uni.Ability):
                    if item.is_abstract:
                        continue
                    if (
                        isinstance(item.name_ref, uni.SpecialVarRef)
                        and item.name_ref.name == Tok.KW_POST_INIT
                    ):
                        postinit_method = item

            # Check if postinit needed and not provided.
            if len(post_init_vars) != 0 and (postinit_method is None):
                self.log_error(
                    'Missing "postinit" method required by un initialized attribute(s).',
                    node_override=post_init_vars[0].name_spec,
                )

    # ------------context update methods---------------------------
    def _update_ctx(self, node: uni.UniNode) -> None:
        if isinstance(node, uni.AtomTrailer):
            self._change_atom_trailer_ctx(node)
        elif isinstance(node, uni.AstSymbolNode):
            node.sym_tab.update_py_ctx_for_def(node)
        else:
            self.log_error(f"Invalid target for context update: {type(node).__name__}")

    def enter_has_var(self, node: uni.HasVar) -> None:
        if isinstance(node.parent, uni.ArchHas):
            node.sym_tab.update_py_ctx_for_def(node)
        else:
            self.ice("HasVar should be under ArchHas")

    def enter_param_var(self, node: uni.ParamVar) -> None:
        node.sym_tab.update_py_ctx_for_def(node)

    def enter_assignment(self, node: uni.Assignment) -> None:
        for target in node.target:
            self._update_ctx(target)

    def enter_in_for_stmt(self, node: uni.InForStmt) -> None:
        self._update_ctx(node.target)

    def enter_expr_as_item(self, node: uni.ExprAsItem) -> None:
        if node.alias:
            self._update_ctx(node.alias)

    def enter_inner_compr(self, node: uni.InnerCompr) -> None:
        self._update_ctx(node.target)

    # ----------------------- Utilities -------------------------

    def _change_atom_trailer_ctx(self, node: uni.AtomTrailer) -> None:
        """Mark final element in trailer chain as a Store context."""
        last = node.right
        if isinstance(last, uni.AtomExpr):
            last.name_spec.py_ctx_func = ast3.Store
            if isinstance(last.name_spec, uni.AstSymbolNode):
                last.name_spec.py_ctx_func = ast3.Store



from .introspection import IntrospectionEndpoint as IntrospectionEndpoint
from .models import IntrospectionToken as IntrospectionToken
from .token_validator import IntrospectTokenValidator as IntrospectTokenValidator

__all__ = ["IntrospectionEndpoint", "IntrospectionToken", "IntrospectTokenValidator"]



from __future__ import annotations

import ctypes
from typing import Type
from typing_extensions import assert_type

assert_type(ctypes.POINTER(None), Type[ctypes.c_void_p])
assert_type(ctypes.POINTER(ctypes.c_int), Type[ctypes._Pointer[ctypes.c_int]])



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

import builtins
import typing

import google.protobuf.descriptor
import google.protobuf.message
import tensorflow.core.framework.versions_pb2

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class FingerprintDef(google.protobuf.message.Message):
    """Protocol buffer representing a SavedModel Fingerprint.

    If there are multiple MetaGraphDefs in the SavedModel, the FingerprintDef
    corresponds to the first one.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    SAVED_MODEL_CHECKSUM_FIELD_NUMBER: builtins.int
    GRAPH_DEF_PROGRAM_HASH_FIELD_NUMBER: builtins.int
    SIGNATURE_DEF_HASH_FIELD_NUMBER: builtins.int
    SAVED_OBJECT_GRAPH_HASH_FIELD_NUMBER: builtins.int
    CHECKPOINT_HASH_FIELD_NUMBER: builtins.int
    VERSION_FIELD_NUMBER: builtins.int
    saved_model_checksum: builtins.int
    """Hash of the saved_model.pb, referred to as a "checksum"."""
    graph_def_program_hash: builtins.int
    """Hash of regularized graph_def."""
    signature_def_hash: builtins.int
    """Hash of the regularized (sorted) SignatureDefs."""
    saved_object_graph_hash: builtins.int
    """Hash of the regularized SavedObjectGraph."""
    checkpoint_hash: builtins.int
    """Hash of the checkpoint."""
    @property
    def version(self) -> tensorflow.core.framework.versions_pb2.VersionDef:
        """Version specification of the fingerprint.
        TODO(b/290068219): add USM version when GA
        """

    def __init__(
        self,
        *,
        saved_model_checksum: builtins.int | None = ...,
        graph_def_program_hash: builtins.int | None = ...,
        signature_def_hash: builtins.int | None = ...,
        saved_object_graph_hash: builtins.int | None = ...,
        checkpoint_hash: builtins.int | None = ...,
        version: tensorflow.core.framework.versions_pb2.VersionDef | None = ...,
    ) -> None: ...
    def HasField(self, field_name: typing.Literal["version", b"version"]) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing.Literal[
            "checkpoint_hash",
            b"checkpoint_hash",
            "graph_def_program_hash",
            b"graph_def_program_hash",
            "saved_model_checksum",
            b"saved_model_checksum",
            "saved_object_graph_hash",
            b"saved_object_graph_hash",
            "signature_def_hash",
            b"signature_def_hash",
            "version",
            b"version",
        ],
    ) -> None: ...

global___FingerprintDef = FingerprintDef



# Internal type that is commonly used as a base class
# and some public apis the signature needs it. As type
# is internal exact module it lives in is unstable across
# versions.
class Trackable: ...



import json
from collections.abc import Callable, Generator, Iterator
from typing import Any

from docker._types import JSON

json_decoder: json.JSONDecoder

def stream_as_text(stream: Iterator[str | bytes]) -> Generator[str]: ...
def json_splitter(buffer: str) -> tuple[JSON, str] | None: ...
def json_stream(stream: Iterator[str]) -> Generator[JSON]: ...
def line_splitter(buffer: str, separator: str = "\n") -> tuple[str, str] | None: ...
def split_buffer(
    stream: Iterator[str | bytes], splitter: Callable[[str], tuple[str, str]] | None = None, decoder: Callable[[str], Any] = ...
) -> Generator[Any]: ...



import threading
from _typeshed import SupportsWrite, TraceFunction
from collections.abc import Callable
from pdb import Pdb
from types import FrameType
from typing import Any, Literal, Protocol, TypeVar, type_check_only
from typing_extensions import TypeAlias

from .watch_element import WatchElement

_T = TypeVar("_T")

# Alias used for fields that must always be valid identifiers
# A string `x` counts as a valid identifier if both the following are True
# (1) `x.isidentifier()` evaluates to `True`
# (2) `keyword.iskeyword(x)` evaluates to `False`
_Identifier: TypeAlias = str

class Watch:
    # User-defined callbacks passed to `__call__()` or `config()` set as instance variables have arguments of type `Any` to be
    # compatible with more precisely-annotated signatures.

    custom_printer: Callable[[Any], None] | None
    enable: bool
    file: str | SupportsWrite[str] | None
    pdb: Pdb | None
    pdb_enable: bool
    set_lock: threading.Lock
    stack_limit: int | None
    tracefunc_lock: threading.Lock
    tracefunc_stack: list[TraceFunction | None]
    watch_list: list[WatchElement]

    def __init__(self) -> None: ...
    def __call__(
        self,
        *args: object,
        alias: str = ...,
        callback: Callable[[FrameType, WatchElement, tuple[str, str, int | None]], None] = ...,
        cmp: Callable[[Any, Any], bool] = ...,  # User-defined comparison callback; compares 2 arguments of any type
        copy: Callable[[_T], _T] = ...,
        # User-defined printing callback; writes a string representation of any object to a stream
        custom_printer: Callable[[Any], None] = ...,
        deepcopy: bool = False,
        file: str | SupportsWrite[str] = ...,
        stack_limit: int | None = 5,
        track: Literal["object", "variable"] = ...,
        when: Callable[[Any], bool] = ...,  # User-defined callback for conditional watchpoints
    ) -> None: ...
    def config(
        self,
        *,
        callback: Callable[[FrameType, WatchElement, tuple[str, str, int | None]], None] = ...,
        pdb: Literal[True] = True,
        file: str | SupportsWrite[str] = ...,
        stack_limit: int | None = 5,
        custom_printer: Callable[[Any], None] = ...,  # User-defined printing callback
    ) -> None: ...
    def install(self, func: _Identifier = "watch") -> None: ...
    def restore(self) -> None: ...
    def start_trace(self, frame: FrameType) -> None: ...
    def stop_trace(self, frame: FrameType) -> None: ...
    def tracefunc(self, frame: FrameType, event: str, arg: object) -> _TraceFunc: ...
    def uninstall(self, func: _Identifier = "watch") -> None: ...
    def unwatch(self, *args: object) -> None: ...

@type_check_only
class _TraceFunc(Protocol):
    def __call__(self, frame: FrameType, event: str, arg: object) -> _TraceFunc: ...



from _typeshed import Incomplete
from typing import ClassVar

from authlib.jose.rfc7517 import Key

class AsymmetricKey(Key):
    PUBLIC_KEY_FIELDS: ClassVar[list[str]]
    PRIVATE_KEY_FIELDS: ClassVar[list[str]]
    PRIVATE_KEY_CLS: ClassVar[type | tuple[type, ...]]
    PUBLIC_KEY_CLS: ClassVar[type | tuple[type, ...]]
    SSH_PUBLIC_PREFIX: ClassVar[bytes]
    private_key: Incomplete
    public_key: Incomplete
    def __init__(self, private_key=None, public_key=None, options=None) -> None: ...
    @property
    def public_only(self): ...
    def get_op_key(self, operation): ...
    def get_public_key(self): ...
    def get_private_key(self): ...
    def load_raw_key(self) -> None: ...
    def load_dict_key(self) -> None: ...
    def dumps_private_key(self): ...
    def dumps_public_key(self): ...
    def load_private_key(self): ...
    def load_public_key(self): ...
    def as_dict(self, is_private: bool = False, **params): ...
    def as_key(self, is_private: bool = False): ...
    def as_bytes(self, encoding=None, is_private: bool = False, password=None): ...
    def as_pem(self, is_private: bool = False, password=None): ...
    def as_der(self, is_private: bool = False, password=None): ...
    @classmethod
    def import_dict_key(cls, raw, options=None): ...
    @classmethod
    def import_key(cls, raw, options=None): ...
    @classmethod
    def validate_raw_key(cls, key): ...
    @classmethod
    def generate_key(cls, crv_or_size, options=None, is_private: bool = False) -> AsymmetricKey: ...



from __future__ import annotations

import sys
from collections.abc import Callable, Iterator
from concurrent.futures import Future, ThreadPoolExecutor, as_completed
from typing import Literal
from typing_extensions import assert_type


class Parent: ...


class Child(Parent): ...


def check_as_completed_covariance() -> None:
    with ThreadPoolExecutor() as executor:
        f1 = executor.submit(lambda: Parent())
        f2 = executor.submit(lambda: Child())
        fs: list[Future[Parent] | Future[Child]] = [f1, f2]
        assert_type(as_completed(fs), Iterator[Future[Parent]])
        for future in as_completed(fs):
            assert_type(future.result(), Parent)


def check_future_invariance() -> None:
    def execute_callback(callback: Callable[[], Parent], future: Future[Parent]) -> None:
        future.set_result(callback())

    fut: Future[Child] = Future()
    execute_callback(lambda: Parent(), fut)  # type: ignore
    assert isinstance(fut.result(), Child)


if sys.version_info >= (3, 14):

    def _initializer(x: int) -> None:
        pass

    def check_interpreter_pool_executor() -> None:
        import concurrent.futures.interpreter
        from concurrent.futures import InterpreterPoolExecutor

        with InterpreterPoolExecutor(initializer=_initializer, initargs=(1,)):
            ...

        with InterpreterPoolExecutor(initializer=_initializer, initargs=("x",)):  # type: ignore
            ...

        context = InterpreterPoolExecutor.prepare_context(initializer=_initializer, initargs=(1,))
        worker_context = context[0]()
        assert_type(worker_context, concurrent.futures.interpreter.WorkerContext)
        resolve_task = context[1]
        # Function should enforce that the arguments are correct.
        res = resolve_task(_initializer, 1)
        assert_type(res, tuple[bytes, Literal["function"]])
        # When the function is a script, the arguments should be a string.
        str_res = resolve_task("print('Hello, world!')")
        assert_type(str_res, tuple[bytes, Literal["script"]])
        # When a script is passed, no arguments should be provided.
        resolve_task("print('Hello, world!')", 1)  # type: ignore

        # `WorkerContext.__init__` should accept the result of a  resolved task.
        concurrent.futures.interpreter.WorkerContext(initdata=res)

        # Run should also accept the result of a resolved task.
        worker_context.run(res)

    def check_thread_worker_context() -> None:
        import concurrent.futures.thread

        context = concurrent.futures.thread.WorkerContext.prepare(initializer=_initializer, initargs=(1,))
        worker_context = context[0]()
        assert_type(worker_context, concurrent.futures.thread.WorkerContext)
        resolve_task = context[1]
        res = resolve_task(_initializer, (1,), {"test": 1})
        assert_type(res[1], tuple[int])
        assert_type(res[2], dict[str, int])



from asyncio import ReadTransport
from collections.abc import Awaitable, Callable, Iterable
from re import Match, Pattern
from typing import IO, AnyStr, Generic, Literal, Protocol, TextIO, overload, type_check_only
from typing_extensions import TypeAlias

from ._async import PatternWaiter
from .exceptions import EOF, TIMEOUT
from .expect import searcher_re, searcher_string

PY3: bool
text_type: type

class _NullCoder:
    @staticmethod
    def encode(b: str, final: bool = False): ...
    @staticmethod
    def decode(b: str, final: bool = False): ...

@type_check_only
class _Logfile(Protocol):
    def write(self, s, /) -> object: ...
    def flush(self) -> object: ...

_ErrorPattern: TypeAlias = type[EOF | TIMEOUT]
_InputStringPattern: TypeAlias = str | bytes | _ErrorPattern
_InputRePattern: TypeAlias = Pattern[str] | Pattern[bytes] | _InputStringPattern
_CompiledStringPattern: TypeAlias = AnyStr | _ErrorPattern
_CompiledRePattern: TypeAlias = Pattern[AnyStr] | _ErrorPattern
_Searcher: TypeAlias = searcher_string[AnyStr] | searcher_re[AnyStr]

class SpawnBase(Generic[AnyStr]):
    encoding: str | None
    pid: int | None
    flag_eof: bool
    stdin: TextIO
    stdout: TextIO
    stderr: TextIO
    searcher: None
    ignorecase: bool
    before: AnyStr | None
    after: _CompiledStringPattern[AnyStr] | None
    match: AnyStr | Match[AnyStr] | _ErrorPattern | None
    match_index: int | None
    terminated: bool
    exitstatus: int | None
    signalstatus: int | None
    status: int | None
    child_fd: int
    timeout: float | None
    delimiter: type[EOF]
    logfile: _Logfile | None
    logfile_read: _Logfile | None
    logfile_send: _Logfile | None
    maxread: int
    searchwindowsize: int | None
    delaybeforesend: float | None
    delayafterclose: float
    delayafterterminate: float
    delayafterread: float | None
    softspace: bool
    name: str
    closed: bool
    codec_errors: str
    string_type: type[AnyStr]
    buffer_type: IO[AnyStr]
    crlf: AnyStr
    allowed_string_types: tuple[type, ...]
    linesep: AnyStr
    write_to_stdout: Callable[[AnyStr], int]
    async_pw_transport: tuple[PatternWaiter[AnyStr], ReadTransport] | None
    def __init__(
        self,
        timeout: float | None = 30,
        maxread: int = 2000,
        searchwindowsize: int | None = None,
        logfile: _Logfile | None = None,
        encoding: str | None = None,
        codec_errors: str = "strict",
    ) -> None: ...
    @property
    def buffer(self) -> AnyStr: ...
    @buffer.setter
    def buffer(self, value: AnyStr) -> None: ...
    def read_nonblocking(self, size: int = 1, timeout: float | None = None) -> AnyStr: ...
    def compile_pattern_list(self, patterns: _InputRePattern | list[_InputRePattern]) -> list[_CompiledRePattern[AnyStr]]: ...
    @overload
    def expect(
        self,
        pattern: _InputRePattern | list[_InputRePattern],
        timeout: float | None = -1,
        searchwindowsize: int | None = -1,
        async_: Literal[False] = False,
    ) -> int: ...
    @overload
    def expect(
        self,
        pattern: _InputRePattern | list[_InputRePattern],
        timeout: float | None = -1,
        searchwindowsize: int | None = -1,
        *,
        async_: Literal[True],
    ) -> Awaitable[int]: ...
    @overload
    def expect_list(
        self,
        pattern_list: list[_CompiledRePattern[AnyStr]],
        timeout: float | None = -1,
        searchwindowsize: int | None = -1,
        async_: Literal[False] = False,
    ) -> int: ...
    @overload
    def expect_list(
        self,
        pattern_list: list[_CompiledRePattern[AnyStr]],
        timeout: float | None = -1,
        searchwindowsize: int | None = -1,
        *,
        async_: Literal[True],
    ) -> Awaitable[int]: ...
    @overload
    def expect_exact(
        self,
        pattern_list: _InputStringPattern | Iterable[_InputStringPattern],
        timeout: float | None = -1,
        searchwindowsize: int | None = -1,
        async_: Literal[False] = False,
    ) -> int: ...
    @overload
    def expect_exact(
        self,
        pattern_list: _InputStringPattern | Iterable[_InputStringPattern],
        timeout: float | None = -1,
        searchwindowsize: int | None = -1,
        *,
        async_: Literal[True],
    ) -> Awaitable[int]: ...
    def expect_loop(self, searcher: _Searcher[AnyStr], timeout: float | None = -1, searchwindowsize: int | None = -1) -> int: ...
    def read(self, size: int = -1) -> AnyStr: ...
    def readline(self, size: int = -1) -> AnyStr: ...
    def __iter__(self): ...
    def readlines(self, sizehint: int = -1) -> list[AnyStr]: ...
    def fileno(self) -> int: ...
    def flush(self) -> None: ...
    def isatty(self) -> bool: ...
    def __enter__(self): ...
    def __exit__(self, etype, evalue, tb) -> None: ...



from networkx.classes.graph import Graph, _Node
from networkx.utils.backends import _dispatchable

__all__ = [
    "margulis_gabber_galil_graph",
    "chordal_cycle_graph",
    "paley_graph",
    "maybe_regular_expander",
    "is_regular_expander",
    "random_regular_expander_graph",
]

@_dispatchable
def margulis_gabber_galil_graph(n, create_using=None): ...
@_dispatchable
def chordal_cycle_graph(p, create_using=None): ...
@_dispatchable
def paley_graph(p, create_using=None): ...
@_dispatchable
def maybe_regular_expander(n, d, *, create_using=None, max_tries=100, seed=None): ...
@_dispatchable
def is_regular_expander(G: Graph[_Node], *, epsilon: float = 0) -> bool: ...
@_dispatchable
def random_regular_expander_graph(n, d, *, epsilon=0, create_using=None, max_tries=100, seed=None): ...



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
Protocol messages for describing features for machine learning model
training or inference.

There are three base Feature types:
  - bytes
  - float
  - int64

A Feature contains Lists which may hold zero or more values.  These
lists are the base values BytesList, FloatList, Int64List.

Features are organized into categories by name.  The Features message
contains the mapping from name to Feature.

Example Features for a movie recommendation application:
  feature {
    key: "age"
    value { float_list {
      value: 29.0
    }}
  }
  feature {
    key: "movie"
    value { bytes_list {
      value: "The Shawshank Redemption"
      value: "Fight Club"
    }}
  }
  feature {
    key: "movie_ratings"
    value { float_list {
      value: 9.0
      value: 9.7
    }}
  }
  feature {
    key: "suggestion"
    value { bytes_list {
      value: "Inception"
    }}
  }
  feature {
    key: "suggestion_purchased"
    value { int64_list {
      value: 1
    }}
  }
  feature {
    key: "purchase_price"
    value { float_list {
      value: 9.99
    }}
  }
"""

import builtins
import collections.abc
import typing

import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class BytesList(google.protobuf.message.Message):
    """LINT.IfChange
    Containers to hold repeated fundamental values.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VALUE_FIELD_NUMBER: builtins.int
    @property
    def value(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.bytes]: ...
    def __init__(self, *, value: collections.abc.Iterable[builtins.bytes] | None = ...) -> None: ...
    def ClearField(self, field_name: typing.Literal["value", b"value"]) -> None: ...

global___BytesList = BytesList

@typing.final
class FloatList(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VALUE_FIELD_NUMBER: builtins.int
    @property
    def value(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.float]: ...
    def __init__(self, *, value: collections.abc.Iterable[builtins.float] | None = ...) -> None: ...
    def ClearField(self, field_name: typing.Literal["value", b"value"]) -> None: ...

global___FloatList = FloatList

@typing.final
class Int64List(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    VALUE_FIELD_NUMBER: builtins.int
    @property
    def value(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.int]: ...
    def __init__(self, *, value: collections.abc.Iterable[builtins.int] | None = ...) -> None: ...
    def ClearField(self, field_name: typing.Literal["value", b"value"]) -> None: ...

global___Int64List = Int64List

@typing.final
class Feature(google.protobuf.message.Message):
    """Containers for non-sequential data."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    BYTES_LIST_FIELD_NUMBER: builtins.int
    FLOAT_LIST_FIELD_NUMBER: builtins.int
    INT64_LIST_FIELD_NUMBER: builtins.int
    @property
    def bytes_list(self) -> global___BytesList: ...
    @property
    def float_list(self) -> global___FloatList: ...
    @property
    def int64_list(self) -> global___Int64List: ...
    def __init__(
        self,
        *,
        bytes_list: global___BytesList | None = ...,
        float_list: global___FloatList | None = ...,
        int64_list: global___Int64List | None = ...,
    ) -> None: ...
    def HasField(
        self,
        field_name: typing.Literal[
            "bytes_list", b"bytes_list", "float_list", b"float_list", "int64_list", b"int64_list", "kind", b"kind"
        ],
    ) -> builtins.bool: ...
    def ClearField(
        self,
        field_name: typing.Literal[
            "bytes_list", b"bytes_list", "float_list", b"float_list", "int64_list", b"int64_list", "kind", b"kind"
        ],
    ) -> None: ...
    def WhichOneof(
        self, oneof_group: typing.Literal["kind", b"kind"]
    ) -> typing.Literal["bytes_list", "float_list", "int64_list"] | None: ...

global___Feature = Feature

@typing.final
class Features(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class FeatureEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        @property
        def value(self) -> global___Feature: ...
        def __init__(self, *, key: builtins.str | None = ..., value: global___Feature | None = ...) -> None: ...
        def HasField(self, field_name: typing.Literal["value", b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> None: ...

    FEATURE_FIELD_NUMBER: builtins.int
    @property
    def feature(self) -> google.protobuf.internal.containers.MessageMap[builtins.str, global___Feature]:
        """Map from feature name to feature."""

    def __init__(self, *, feature: collections.abc.Mapping[builtins.str, global___Feature] | None = ...) -> None: ...
    def ClearField(self, field_name: typing.Literal["feature", b"feature"]) -> None: ...

global___Features = Features

@typing.final
class FeatureList(google.protobuf.message.Message):
    """Containers for sequential data.

    A FeatureList contains lists of Features.  These may hold zero or more
    Feature values.

    FeatureLists are organized into categories by name.  The FeatureLists message
    contains the mapping from name to FeatureList.
    """

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    FEATURE_FIELD_NUMBER: builtins.int
    @property
    def feature(self) -> google.protobuf.internal.containers.RepeatedCompositeFieldContainer[global___Feature]: ...
    def __init__(self, *, feature: collections.abc.Iterable[global___Feature] | None = ...) -> None: ...
    def ClearField(self, field_name: typing.Literal["feature", b"feature"]) -> None: ...

global___FeatureList = FeatureList

@typing.final
class FeatureLists(google.protobuf.message.Message):
    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    @typing.final
    class FeatureListEntry(google.protobuf.message.Message):
        DESCRIPTOR: google.protobuf.descriptor.Descriptor

        KEY_FIELD_NUMBER: builtins.int
        VALUE_FIELD_NUMBER: builtins.int
        key: builtins.str
        @property
        def value(self) -> global___FeatureList: ...
        def __init__(self, *, key: builtins.str | None = ..., value: global___FeatureList | None = ...) -> None: ...
        def HasField(self, field_name: typing.Literal["value", b"value"]) -> builtins.bool: ...
        def ClearField(self, field_name: typing.Literal["key", b"key", "value", b"value"]) -> None: ...

    FEATURE_LIST_FIELD_NUMBER: builtins.int
    @property
    def feature_list(self) -> google.protobuf.internal.containers.MessageMap[builtins.str, global___FeatureList]:
        """Map from feature name to feature list."""

    def __init__(self, *, feature_list: collections.abc.Mapping[builtins.str, global___FeatureList] | None = ...) -> None: ...
    def ClearField(self, field_name: typing.Literal["feature_list", b"feature_list"]) -> None: ...

global___FeatureLists = FeatureLists



from logging import Logger
from typing import Final

from .context import Context

log: Logger
LAMBDA_TRACE_HEADER_KEY: Final = "_X_AMZN_TRACE_ID"
LAMBDA_TASK_ROOT_KEY: Final = "LAMBDA_TASK_ROOT"
TOUCH_FILE_DIR: Final = "/tmp/.aws-xray/"
TOUCH_FILE_PATH: Final = "/tmp/.aws-xray/initialized"

def check_in_lambda() -> LambdaContext | None: ...

class LambdaContext(Context):
    def __init__(self) -> None: ...
    @property  # type: ignore[override]
    def context_missing(self) -> None: ...
    @context_missing.setter
    def context_missing(self, value: str) -> None: ...



from _typeshed import Incomplete
from collections.abc import Callable, Iterable

from networkx.classes.graph import Graph, _Node
from networkx.utils.backends import _dispatchable

__all__ = ["min_edge_cover", "is_edge_cover"]

@_dispatchable
def min_edge_cover(G: Graph[_Node], matching_algorithm: Callable[..., Incomplete] | None = None) -> set[Incomplete]: ...
@_dispatchable
def is_edge_cover(G: Graph[_Node], cover: Iterable[Iterable[Incomplete]]) -> bool: ...



import ssl
from _socket import _Address
from _typeshed import StrOrBytesPath
from collections.abc import Callable, Iterable, Mapping
from socket import AddressFamily, SocketKind
from typing import Any
from typing_extensions import TypeAlias

from ..socks import sockssocket
from ..utils.networking import HTTPHeaderDict
from .common import Request, RequestHandler, Response

def ssl_load_certs(context: ssl.SSLContext, use_certifi: bool = True) -> None: ...
def ssl_load_windows_store_certs(ssl_context: ssl.SSLContext, storename: str) -> None: ...
def make_socks_proxy_opts(socks_proxy: str) -> dict[str, Any]: ...
def get_redirect_method(method: str, status: int) -> str: ...
def make_ssl_context(
    verify: bool = True,
    client_certificate: StrOrBytesPath | None = None,
    client_certificate_key: StrOrBytesPath | None = None,
    client_certificate_password: Callable[[], str | bytes | bytearray] | str | bytes | bytearray | None = None,
    legacy_support: bool = False,
    use_certifi: bool = True,
) -> ssl.SSLContext: ...

class InstanceStoreMixin:
    def __init__(self, **kwargs: Any) -> None: ...  # Passed to non-existent parent so MRO works.

def add_accept_encoding_header(headers: HTTPHeaderDict, supported_encodings: Iterable[str]) -> None: ...
def wrap_request_errors(
    func: Callable[[RequestHandler, Request], Response | None],
) -> Callable[[RequestHandler, Request], None]: ...

_IPAddress: TypeAlias = tuple[
    AddressFamily, SocketKind, int, str, tuple[str, int] | tuple[str, int, int, int] | tuple[int, bytes]
]

def create_socks_proxy_socket(
    dest_addr: _Address, proxy_args: Mapping[str, Any], proxy_ip_addr: _IPAddress, timeout: float | None, source_address: _Address
) -> sockssocket: ...
def create_connection(
    address: tuple[str, int],
    timeout: int = ...,
    source_address: _Address | None = None,
    *,
    _create_socket_func: Callable[[_IPAddress, int, _Address], sockssocket] = ...,
) -> sockssocket: ...



from collections.abc import Iterable
from typing import overload
from typing_extensions import Self, TypeAlias

from .._typing import ArrayLikeSeq
from .base import BaseGeometry
from .collection import GeometryCollection

__all__ = ["Point"]

_PointLike: TypeAlias = Point | Iterable[float] | ArrayLikeSeq[float]

class Point(BaseGeometry):
    __slots__: list[str] = []
    @overload  # no args: empty point
    def __new__(self) -> Self: ...
    @overload  # one arg: (x, y[, z]) tuple or a Point instance
    def __new__(self, coords: _PointLike, /) -> Self: ...
    @overload  # two args: (x, y) tuple
    def __new__(self, x: float, y: float, /) -> Self: ...
    @overload  # three args: (x, y, z) tuple
    def __new__(self, x: float, y: float, z: float, /) -> Self: ...
    @property
    def x(self) -> float: ...
    @property
    def y(self) -> float: ...
    @property
    def z(self) -> float: ...
    @property
    def m(self) -> float: ...
    def svg(self, scale_factor: float = 1.0, fill_color: str | None = None, opacity: float | None = None) -> str: ...  # type: ignore[override]
    # more precise base overrides
    @property
    def boundary(self) -> GeometryCollection: ...  # empty geometry collection
    @property
    def convex_hull(self) -> Point: ...
    @property
    def envelope(self) -> Point: ...
    @property
    def oriented_envelope(self) -> Point: ...
    @property
    def minimum_rotated_rectangle(self) -> Point: ...



from collections.abc import Mapping
from typing import Any, SupportsInt
from typing_extensions import deprecated

from jwcrypto.common import JWException, JWKeyNotFound
from jwcrypto.jwk import JWK, JWKSet

JWTClaimsRegistry: Mapping[str, str]
JWT_expect_type: bool

class JWTExpired(JWException):
    def __init__(self, message: str | None = None, exception: BaseException | None = None) -> None: ...

class JWTNotYetValid(JWException):
    def __init__(self, message: str | None = None, exception: BaseException | None = None) -> None: ...

class JWTMissingClaim(JWException):
    def __init__(self, message: str | None = None, exception: BaseException | None = None) -> None: ...

class JWTInvalidClaimValue(JWException):
    def __init__(self, message: str | None = None, exception: BaseException | None = None) -> None: ...

class JWTInvalidClaimFormat(JWException):
    def __init__(self, message: str | None = None, exception: BaseException | None = None) -> None: ...

@deprecated("")
class JWTMissingKeyID(JWException):
    def __init__(self, message: str | None = None, exception: BaseException | None = None) -> None: ...

class JWTMissingKey(JWKeyNotFound):
    def __init__(self, message: str | None = None, exception: BaseException | None = None) -> None: ...

class JWT:
    deserializelog: list[str] | None
    def __init__(
        self,
        header: dict[str, Any] | str | None = None,
        claims: dict[str, Any] | str | None = None,
        jwt=None,
        key: JWK | JWKSet | None = None,
        algs=None,
        default_claims=None,
        check_claims=None,
        expected_type=None,
    ) -> None: ...
    @property
    def header(self) -> str: ...
    @header.setter
    def header(self, h: dict[str, Any] | str) -> None: ...
    @property
    def claims(self) -> str: ...
    @claims.setter
    def claims(self, data: str) -> None: ...
    @property
    def token(self): ...
    @token.setter
    def token(self, t) -> None: ...
    @property
    def leeway(self) -> int: ...
    @leeway.setter
    def leeway(self, lwy: SupportsInt) -> None: ...
    @property
    def validity(self) -> int: ...
    @validity.setter
    def validity(self, v: SupportsInt) -> None: ...
    @property
    def expected_type(self): ...
    @expected_type.setter
    def expected_type(self, v) -> None: ...
    def norm_typ(self, val): ...
    def make_signed_token(self, key: JWK) -> None: ...
    def make_encrypted_token(self, key: JWK) -> None: ...
    def validate(self, key: JWK | JWKSet) -> None: ...
    def deserialize(self, jwt, key=None) -> None: ...
    def serialize(self, compact: bool = True) -> str: ...
    @classmethod
    def from_jose_token(cls, token): ...
    def __eq__(self, other: object) -> bool: ...



import sys
from typing import Any
from typing_extensions import TypeAlias

from gevent._fileobjectcommon import FileObjectBlock as FileObjectBlock, FileObjectThread as FileObjectThread

if sys.platform != "win32":
    import io
    from _typeshed import (
        FileDescriptorOrPath,
        OpenBinaryMode,
        OpenBinaryModeReading,
        OpenBinaryModeUpdating,
        OpenBinaryModeWriting,
        OpenTextMode,
    )
    from typing import IO, AnyStr, Literal, overload

    from gevent._fileobjectcommon import _IOT, FileObjectBase

    # this is implemented in _fileobjectposix and technically uses an undocumented subclass
    # of RawIOBase, but the interface is the same, so it doesn't seem worth it to add
    # annotations for it. _fileobjectcommon was barely worth it due to the common base class
    # of all three FileObject types
    class FileObjectPosix(FileObjectBase[_IOT, AnyStr]):
        default_bufsize = io.DEFAULT_BUFFER_SIZE
        fileio: io.RawIOBase
        # Text mode: always binds a TextIOWrapper
        @overload
        def __init__(
            self: FileObjectPosix[io.TextIOWrapper, str],
            fobj: FileDescriptorOrPath,
            mode: OpenTextMode = "r",
            bufsize: int | None = None,
            close: bool | None = None,
            encoding: str | None = None,
            errors: str | None = None,
            newline: str | None = None,
            buffering: int | None = None,
            closefd: bool | None = None,
            atomic_write: bool = False,
        ) -> None: ...

        # Unbuffered binary mode: binds a FileIO
        @overload
        def __init__(
            self: FileObjectPosix[io.FileIO, bytes],
            fobj: FileDescriptorOrPath,
            mode: OpenBinaryMode,
            bufsize: Literal[0],
            close: bool | None = None,
            encoding: str | None = None,
            errors: str | None = None,
            newline: str | None = None,
            buffering: Literal[0] | None = None,
            closefd: bool | None = None,
            atomic_write: bool = False,
        ) -> None: ...
        @overload
        def __init__(
            self: FileObjectPosix[io.FileIO, bytes],
            fobj: FileDescriptorOrPath,
            mode: OpenBinaryMode,
            bufsize: Literal[0] | None = None,
            close: bool | None = None,
            encoding: str | None = None,
            errors: str | None = None,
            newline: str | None = None,
            *,
            buffering: Literal[0],
            closefd: bool | None = None,
            atomic_write: bool = False,
        ) -> None: ...

        # Buffering is on: return BufferedRandom, BufferedReader, or BufferedWriter
        @overload
        def __init__(
            self: FileObjectPosix[io.BufferedRandom, bytes],
            fobj: FileDescriptorOrPath,
            mode: OpenBinaryModeUpdating,
            bufsize: Literal[-1, 1] | None = None,
            close: bool | None = None,
            encoding: str | None = None,
            errors: str | None = None,
            newline: str | None = None,
            buffering: Literal[-1, 1] | None = None,
            closefd: bool | None = None,
            atomic_write: bool = False,
        ) -> None: ...
        @overload
        def __init__(
            self: FileObjectPosix[io.BufferedWriter, bytes],
            fobj: FileDescriptorOrPath,
            mode: OpenBinaryModeWriting,
            bufsize: Literal[-1, 1] | None = None,
            close: bool | None = None,
            encoding: str | None = None,
            errors: str | None = None,
            newline: str | None = None,
            buffering: Literal[-1, 1] | None = None,
            closefd: bool | None = None,
            atomic_write: bool = False,
        ) -> None: ...
        @overload
        def __init__(
            self: FileObjectPosix[io.BufferedReader, bytes],
            fobj: FileDescriptorOrPath,
            mode: OpenBinaryModeReading,
            bufsize: Literal[-1, 1] | None = None,
            close: bool | None = None,
            encoding: str | None = None,
            errors: str | None = None,
            newline: str | None = None,
            buffering: Literal[-1, 1] | None = None,
            closefd: bool | None = None,
            atomic_write: bool = False,
        ) -> None: ...

        # Buffering cannot be determined: fall back to BinaryIO
        @overload
        def __init__(
            self: FileObjectPosix[IO[bytes], bytes],
            fobj: FileDescriptorOrPath,
            mode: OpenBinaryMode,
            bufsize: int | None = None,
            close: bool | None = None,
            encoding: str | None = None,
            errors: str | None = None,
            newline: str | None = None,
            buffering: int | None = None,
            closefd: bool | None = None,
            atomic_write: bool = False,
        ) -> None: ...

        # Fallback if mode is not specified
        @overload
        def __init__(
            self: FileObjectPosix[IO[Any], Any],
            fobj: FileDescriptorOrPath,
            mode: str,
            bufsize: int | None = None,
            close: bool | None = None,
            encoding: str | None = None,
            errors: str | None = None,
            newline: str | None = None,
            buffering: int | None = None,
            closefd: bool | None = None,
            atomic_write: bool = False,
        ) -> None: ...

    _FileObjectType: TypeAlias = type[FileObjectPosix[Any, Any] | FileObjectBlock[Any, Any] | FileObjectThread[Any, Any]]
    __all__ = ["FileObjectPosix", "FileObjectThread", "FileObjectBlock", "FileObject"]
else:
    _FileObjectType: TypeAlias = type[FileObjectBlock[Any, Any] | FileObjectThread[Any, Any]]
    __all__ = ["FileObjectThread", "FileObjectBlock", "FileObject"]

FileObject: _FileObjectType



from abc import ABC, abstractmethod
from typing import Literal

class Transition(ABC):
    @abstractmethod
    def serialize(self) -> str: ...

class SplitTransition(Transition):
    dimension: Literal["H", "V"]
    direction: Literal["I", "O"]
    def __init__(self, dimension: Literal["H", "V"], direction: Literal["I", "O"]) -> None: ...
    def serialize(self) -> str: ...

class BlindsTransition(Transition):
    dimension: Literal["H", "V"]
    def __init__(self, dimension: Literal["H", "V"]) -> None: ...
    def serialize(self) -> str: ...

class BoxTransition(Transition):
    direction: Literal["I", "O"]
    def __init__(self, direction: Literal["I", "O"]) -> None: ...
    def serialize(self) -> str: ...

class WipeTransition(Transition):
    direction: Literal[0, 90, 180, 270]
    def __init__(self, direction: Literal[0, 90, 180, 270]) -> None: ...
    def serialize(self) -> str: ...

class DissolveTransition(Transition):
    def serialize(self) -> str: ...

class GlitterTransition(Transition):
    direction: Literal[0, 270, 315]
    def __init__(self, direction: Literal[0, 270, 315]) -> None: ...
    def serialize(self) -> str: ...

class FlyTransition(Transition):
    dimension: Literal["H", "V"]
    direction: Literal[0, 270] | None
    def __init__(self, dimension: Literal["H", "V"], direction: Literal[0, 270] | None = None) -> None: ...
    def serialize(self) -> str: ...

class PushTransition(Transition):
    direction: Literal[0, 270]
    def __init__(self, direction: Literal[0, 270]) -> None: ...
    def serialize(self) -> str: ...

class CoverTransition(Transition):
    direction: Literal[0, 270]
    def __init__(self, direction: Literal[0, 270]) -> None: ...
    def serialize(self) -> str: ...

class UncoverTransition(Transition):
    direction: Literal[0, 270]
    def __init__(self, direction: Literal[0, 270]) -> None: ...
    def serialize(self) -> str: ...

class FadeTransition(Transition):
    def serialize(self) -> str: ...



from _typeshed import Incomplete
from collections.abc import Generator

from networkx.classes.graph import Graph, _Node
from networkx.utils.backends import _dispatchable

__all__ = [
    "generate_edgelist",
    "write_edgelist",
    "parse_edgelist",
    "read_edgelist",
    "read_weighted_edgelist",
    "write_weighted_edgelist",
]

def generate_edgelist(G: Graph[_Node], delimiter: str = " ", data: bool = True) -> Generator[Incomplete, None, None]: ...
def write_edgelist(
    G: Graph[_Node], path, comments: str = "#", delimiter: str = " ", data: bool = True, encoding: str = "utf-8"
) -> None: ...
@_dispatchable
def parse_edgelist(lines, comments: str = "#", delimiter=None, create_using=None, nodetype=None, data: bool = True): ...
@_dispatchable
def read_edgelist(
    path,
    comments: str = "#",
    delimiter=None,
    create_using=None,
    nodetype=None,
    data: bool = True,
    edgetype=None,
    encoding: str = "utf-8",
): ...
def write_weighted_edgelist(
    G: Graph[_Node], path, comments: str = "#", delimiter: str = " ", encoding: str = "utf-8"
) -> None: ...
@_dispatchable
def read_weighted_edgelist(
    path, comments: str = "#", delimiter=None, create_using=None, nodetype=None, encoding: str = "utf-8"
): ...



from _typeshed import Incomplete

from authlib.oauth2 import OAuth2Error

__all__ = ["InvalidTokenError", "InsufficientScopeError"]

class InvalidTokenError(OAuth2Error):
    error: str
    description: str
    status_code: int
    realm: Incomplete
    extra_attributes: Incomplete
    def __init__(self, description=None, uri=None, status_code=None, state=None, realm=None, **extra_attributes) -> None: ...
    def get_headers(self): ...

class InsufficientScopeError(OAuth2Error):
    error: str
    description: str
    status_code: int



from collections.abc import Callable
from typing import Any

from click.decorators import _AnyCallable

from .core import Shell

def shell(name: str | None = None, **attrs: Any) -> Callable[[_AnyCallable], Shell]: ...



from collections.abc import Callable, Iterable, Mapping
from typing import Final, Literal, Protocol, TypedDict, TypeVar, type_check_only
from typing_extensions import Unpack

from _win32typing import PyHKEY, PyIID

_T = TypeVar("_T", PyHKEY, int)

@type_check_only
class _RegisterClass(Protocol):
    _reg_clsid_: PyIID

@type_check_only
class _RegisterFlag(TypedDict, total=False):
    quiet: bool
    debug: bool
    finalize_register: Callable[[], None]

@type_check_only
class _UnregisterFlag(TypedDict, total=False):
    quiet: bool
    finalize_unregister: Callable[[], None]

@type_check_only
class _ElevatedFlag(TypedDict, total=False):
    quiet: bool
    unattended: bool
    hwnd: int

@type_check_only
class _CommandFlag(_RegisterFlag, _UnregisterFlag, _ElevatedFlag):  # type: ignore[misc]
    ...

CATID_PythonCOMServer: Final = "{B3EF80D0-68E2-11D0-A689-00C04FD658FF}"

def recurse_delete_key(path: str | None, base: PyHKEY | int = -2147483648) -> None: ...
def RegisterServer(
    clsid: PyIID,
    pythonInstString: str | None = None,
    desc: str | None = None,
    progID: str | None = None,
    verProgID: str | None = None,
    defIcon: str | None = None,
    threadingModel: Literal["apartment", "both", "free", "neutral"] = "both",
    policy: str | None = None,
    catids: list[PyIID] = [],
    other: Mapping[str, str] = {},
    addPyComCat: bool | None = None,
    dispatcher: str | None = None,
    clsctx: int | None = None,
    addnPath: str | None = None,
) -> None: ...
def GetUnregisterServerKeys(
    clsid: PyIID, progID: str | None = None, verProgID: str | None = None, customKeys: Iterable[tuple[str, _T]] | None = None
) -> list[tuple[str, _T | int]]: ...
def UnregisterServer(
    clsid: PyIID,
    progID: str | None = None,
    verProgID: str | None = None,
    customKeys: Iterable[tuple[str, PyHKEY | int]] | None = None,
) -> None: ...
def GetRegisteredServerOption(clsid: PyIID, optionName: str) -> str | None: ...
def RegisterClasses(*classes: type[_RegisterClass], **flags: Unpack[_RegisterFlag]) -> None: ...
def UnregisterClasses(*classes: type[_RegisterClass], **flags: Unpack[_UnregisterFlag]) -> None: ...
def UnregisterInfoClasses(*classes: type[_RegisterClass]) -> list[tuple[str, PyHKEY | int]]: ...
def ReExecuteElevated(flags: _ElevatedFlag) -> None: ...
def UseCommandLine(*classes: type[_RegisterClass], **flags: Unpack[_CommandFlag]) -> list[tuple[str, PyHKEY | int]] | None: ...
def RegisterPyComCategory() -> None: ...



from networkx.classes.digraph import DiGraph
from networkx.classes.graph import _Node
from networkx.utils.backends import _dispatchable

__all__ = ["tree_data", "tree_graph"]

def tree_data(G: DiGraph[_Node], root, ident: str = "id", children: str = "children"): ...
@_dispatchable
def tree_graph(data, ident: str = "id", children: str = "children"): ...



from _typeshed import FileDescriptorOrPath
from os import _Environ
from typing import AnyStr

from .spawnbase import _InputRePattern, _Logfile

def run(
    command: str,
    timeout: float | None = 30,
    withexitstatus: bool = False,
    events: list[tuple[_InputRePattern, AnyStr]] | dict[_InputRePattern, AnyStr] | None = None,
    extra_args: None = None,
    logfile: _Logfile | None = None,
    cwd: FileDescriptorOrPath | None = None,
    env: _Environ[str] | None = None,
    **kwargs,
) -> AnyStr | tuple[AnyStr, int]: ...
def runu(
    command: str,
    timeout: float | None = 30,
    withexitstatus: bool = False,
    events: list[tuple[_InputRePattern, AnyStr]] | dict[_InputRePattern, AnyStr] | None = None,
    extra_args: None = None,
    logfile: _Logfile | None = None,
    cwd: FileDescriptorOrPath | None = None,
    env: _Environ[str] | None = None,
    **kwargs,
) -> AnyStr | tuple[AnyStr, int]: ...



from _typeshed import Incomplete
from typing import TextIO

from antlr4 import *

def serializedATN(): ...

class XPathLexer(Lexer):
    atn: Incomplete
    decisionsToDFA: Incomplete
    TOKEN_REF: int
    RULE_REF: int
    ANYWHERE: int
    ROOT: int
    WILDCARD: int
    BANG: int
    ID: int
    STRING: int
    channelNames: Incomplete
    modeNames: Incomplete
    literalNames: Incomplete
    symbolicNames: Incomplete
    ruleNames: Incomplete
    grammarFileName: str
    def __init__(self, input=None, output: TextIO = ...) -> None: ...
    def action(self, localctx: RuleContext, ruleIndex: int, actionIndex: int): ...
    type: Incomplete
    def ID_action(self, localctx: RuleContext, actionIndex: int): ...



"""
@generated by mypy-protobuf.  Do not edit manually!
isort:skip_file
"""

import builtins
import collections.abc
import typing

import google.protobuf.descriptor
import google.protobuf.internal.containers
import google.protobuf.message
import tensorflow.tsl.protobuf.error_codes_pb2

DESCRIPTOR: google.protobuf.descriptor.FileDescriptor

@typing.final
class QueueRunnerDef(google.protobuf.message.Message):
    """Protocol buffer representing a QueueRunner."""

    DESCRIPTOR: google.protobuf.descriptor.Descriptor

    QUEUE_NAME_FIELD_NUMBER: builtins.int
    ENQUEUE_OP_NAME_FIELD_NUMBER: builtins.int
    CLOSE_OP_NAME_FIELD_NUMBER: builtins.int
    CANCEL_OP_NAME_FIELD_NUMBER: builtins.int
    QUEUE_CLOSED_EXCEPTION_TYPES_FIELD_NUMBER: builtins.int
    queue_name: builtins.str
    """Queue name."""
    close_op_name: builtins.str
    """The operation to run to close the queue."""
    cancel_op_name: builtins.str
    """The operation to run to cancel the queue."""
    @property
    def enqueue_op_name(self) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[builtins.str]:
        """A list of enqueue operations."""

    @property
    def queue_closed_exception_types(
        self,
    ) -> google.protobuf.internal.containers.RepeatedScalarFieldContainer[tensorflow.tsl.protobuf.error_codes_pb2.Code.ValueType]:
        """A list of exception types considered to signal a safely closed queue
        if raised during enqueue operations.
        """

    def __init__(
        self,
        *,
        queue_name: builtins.str | None = ...,
        enqueue_op_name: collections.abc.Iterable[builtins.str] | None = ...,
        close_op_name: builtins.str | None = ...,
        cancel_op_name: builtins.str | None = ...,
        queue_closed_exception_types: (
            collections.abc.Iterable[tensorflow.tsl.protobuf.error_codes_pb2.Code.ValueType] | None
        ) = ...,
    ) -> None: ...
    def ClearField(
        self,
        field_name: typing.Literal[
            "cancel_op_name",
            b"cancel_op_name",
            "close_op_name",
            b"close_op_name",
            "enqueue_op_name",
            b"enqueue_op_name",
            "queue_closed_exception_types",
            b"queue_closed_exception_types",
            "queue_name",
            b"queue_name",
        ],
    ) -> None: ...

global___QueueRunnerDef = QueueRunnerDef



from _typeshed import Incomplete
from collections.abc import Generator

from networkx.classes.graph import Graph, _Node
from networkx.utils.backends import _dispatchable
from numpy.random import RandomState

__all__ = ["louvain_communities", "louvain_partitions"]

@_dispatchable
def louvain_communities(
    G: Graph[_Node],
    weight: str | None = "weight",
    resolution: float | None = 1,
    threshold: float | None = 1e-07,
    max_level: int | None = None,
    seed: int | RandomState | None = None,
) -> list[set[Incomplete]]: ...
@_dispatchable
def louvain_partitions(
    G: Graph[_Node],
    weight: str | None = "weight",
    resolution: float | None = 1,
    threshold: float | None = 1e-07,
    seed: int | RandomState | None = None,
) -> Generator[list[set[Incomplete]], None, None]: ...



from hvac.api.vault_api_base import VaultApiBase

DEFAULT_MOUNT_POINT: str

class Database(VaultApiBase):
    def configure(
        self,
        name,
        plugin_name,
        verify_connection=None,
        allowed_roles=None,
        root_rotation_statements=None,
        mount_point="database",
        *args,
        **kwargs,
    ): ...
    def rotate_root_credentials(self, name, mount_point="database"): ...
    def read_connection(self, name, mount_point="database"): ...
    def list_connections(self, mount_point="database"): ...
    def delete_connection(self, name, mount_point="database"): ...
    def reset_connection(self, name, mount_point="database"): ...
    def create_role(
        self,
        name,
        db_name,
        creation_statements,
        default_ttl=None,
        max_ttl=None,
        revocation_statements=None,
        rollback_statements=None,
        renew_statements=None,
        mount_point="database",
    ): ...
    def create_static_role(
        self, name, db_name, username, rotation_statements, rotation_period: int = 86400, mount_point="database"
    ): ...
    def read_role(self, name, mount_point="database"): ...
    def read_static_role(self, name, mount_point="database"): ...
    def list_roles(self, mount_point="database"): ...
    def list_static_roles(self, mount_point="database"): ...
    def delete_role(self, name, mount_point="database"): ...
    def delete_static_role(self, name, mount_point="database"): ...
    def generate_credentials(self, name, mount_point="database"): ...
    def get_static_credentials(self, name, mount_point="database"): ...
    def rotate_static_role_credentials(self, name, mount_point="database"): ...



def to_jd(year: int, dayofyear: int) -> float: ...
def from_jd(jd: float) -> tuple[int, int]: ...
def from_gregorian(year: int, month: int, day: int) -> tuple[int, int]: ...
def to_gregorian(year: int, dayofyear: int) -> tuple[int, int, int]: ...



from _typeshed import Incomplete
from decimal import Decimal
from typing import Final

from braintree.attribute_getter import AttributeGetter
from braintree.dispute_details import DisputeEvidence, DisputePayPalMessage, DisputeStatusHistory
from braintree.error_result import ErrorResult
from braintree.successful_result import SuccessfulResult
from braintree.transaction_details import TransactionDetails

class Dispute(AttributeGetter):
    class Status:
        Accepted: Final = "accepted"
        AutoAccepted: Final = "auto_accepted"
        Disputed: Final = "disputed"
        Expired: Final = "expired"
        Lost: Final = "lost"
        Open: Final = "open"
        UnderReview: Final = "under_review"
        Won: Final = "won"

    class Reason:
        CancelledRecurringTransaction: Final = "cancelled_recurring_transaction"
        CreditNotProcessed: Final = "credit_not_processed"
        Duplicate: Final = "duplicate"
        Fraud: Final = "fraud"
        General: Final = "general"
        InvalidAccount: Final = "invalid_account"
        NotRecognized: Final = "not_recognized"
        ProductNotReceived: Final = "product_not_received"
        ProductUnsatisfactory: Final = "product_unsatisfactory"
        Retrieval: Final = "retrieval"
        TransactionAmountDiffers: Final = "transaction_amount_differs"

    class Kind:
        Chargeback: Final = "chargeback"
        PreArbitration: Final = "pre_arbitration"
        Retrieval: Final = "retrieval"

    class ChargebackProtectionLevel:
        Effortless: Final = "effortless"
        Standard: Final = "standard"
        NotProtected: Final = "not_protected"

    class PreDisputeProgram:
        NONE: Final = "none"
        VisaRdr: Final = "visa_rdr"

    class ProtectionLevel:
        EffortlessCBP: Final = "Effortless Chargeback Protection tool"
        StandardCBP: Final = "Chargeback Protection tool"
        NoProtection: Final = "No Protection"

    @staticmethod
    def accept(id: str) -> SuccessfulResult | ErrorResult: ...
    @staticmethod
    def add_file_evidence(dispute_id: str, document_upload_id) -> SuccessfulResult | ErrorResult | None: ...
    @staticmethod
    def add_text_evidence(id: str, content_or_request) -> SuccessfulResult | ErrorResult | None: ...
    @staticmethod
    def finalize(id: str) -> SuccessfulResult | ErrorResult: ...
    @staticmethod
    def find(id: str) -> Dispute: ...
    @staticmethod
    def remove_evidence(id: str, evidence_id: str) -> SuccessfulResult | ErrorResult: ...
    @staticmethod
    def search(*query) -> SuccessfulResult: ...
    amount: Decimal | None
    amount_disputed: Decimal | None
    amount_won: Decimal | None
    protection_level: Incomplete
    transaction_details: TransactionDetails
    transaction = transaction_details
    evidence: list[DisputeEvidence] | None
    paypal_messages: list[DisputePayPalMessage] | None
    status_history: list[DisputeStatusHistory] | None
    processor_comments: Incomplete
    forwarded_comments: processor_comments
    def __init__(self, attributes) -> None: ...



# Needed until mypy issues are solved or https://github.com/python/mypy/issues/12358
# pyright: reportUnnecessaryTypeIgnoreComment=false
from __future__ import annotations

from _typeshed import ReadableBuffer
from datetime import date, datetime, time
from typing import Any, List, Literal, Tuple, Union
from typing_extensions import assert_type

from openpyxl.descriptors import Strict
from openpyxl.descriptors.base import (
    Bool,
    Convertible,
    DateTime,
    Descriptor,
    Float,
    Integer,
    Length,
    MatchPattern,
    MinMax,
    NoneSet,
    Set,
    String,
    Typed,
)
from openpyxl.descriptors.serialisable import Serialisable


class WithDescriptors(Serialisable):
    descriptor = Descriptor[str]()

    typed_default = Typed(expected_type=str)
    typed_not_none = Typed(expected_type=str, allow_none=False)
    typed_none = Typed(expected_type=str, allow_none=True)

    set_tuple = Set(values=("a", 1, 0.0))
    set_list = Set(values=["a", 1, 0.0])
    set_tuple_none = Set(values=("a", 1, 0.0, None))

    noneset_tuple = NoneSet(values=("a", 1, 0.0))
    noneset_list = NoneSet(values=["a", 1, 0.0])

    length_tuple = Length[Tuple[str, str]](length=1)  # Can't validate tuple length in a generic manner
    length_list = Length[List[str]](length=1)
    length_invalid = Length[object](length=1)  # type: ignore

    match_pattern_str_default = MatchPattern(pattern="")
    match_pattern_str = MatchPattern(pattern="", allow_none=False)
    match_pattern_str_none = MatchPattern(pattern="", allow_none=True)
    match_pattern_bytes_default = MatchPattern(pattern=b"")
    match_pattern_bytes = MatchPattern(pattern=b"", allow_none=False)
    match_pattern_bytes_none = MatchPattern(pattern=b"", allow_none=True)

    convertible_default = Convertible(expected_type=int)
    convertible_not_none = Convertible(expected_type=int, allow_none=False)
    convertible_none = Convertible(expected_type=int, allow_none=True)

    # NOTE: min and max params are independent of expected_type since int and floats can always be compared together
    minmax_default = MinMax(min=0, max=0)
    minmax_float = MinMax(min=0, max=0, expected_type=float, allow_none=False)
    minmax_float_none = MinMax(min=0, max=0, expected_type=float, allow_none=True)
    minmax_int = MinMax(min=0.0, max=0.0, expected_type=int, allow_none=False)
    minmax_int_none = MinMax(min=0.0, max=0.0, expected_type=int, allow_none=True)

    bool_default = Bool()
    bool_not_none = Bool(allow_none=False)
    bool_none = Bool(allow_none=True)

    datetime_default = DateTime()
    datetime_not_none = DateTime(allow_none=False)
    datetime_none = DateTime(allow_none=True)

    string_default = String()
    string_not_none = String(allow_none=False)
    string_none = String(allow_none=True)

    float_default = Float()
    float_not_none = Float(allow_none=False)
    float_none = Float(allow_none=True)

    integer_default = Integer()
    integer_not_none = Integer(allow_none=False)
    integer_none = Integer(allow_none=True)

    # Test inferred annotation
    assert_type(descriptor, Descriptor[str])

    assert_type(typed_default, Typed[str, Literal[False]])
    assert_type(typed_not_none, Typed[str, Literal[False]])
    assert_type(typed_none, Typed[str, Literal[True]])

    assert_type(set_tuple, Set[Union[Literal["a", 1], float]])  # type: ignore[assert-type]  # False-positive in mypy
    assert_type(set_list, Set[Union[str, int, float]])  # type: ignore[assert-type]  # False-positive in mypy # Literals are simplified in non-tuples
    assert_type(set_tuple_none, Set[Union[Literal["a", 1, None], float]])  # type: ignore[assert-type]  # False-positive in mypy

    assert_type(noneset_tuple, NoneSet[Union[Literal["a", 1], float]])  # type: ignore[assert-type]  # False-positive in mypy
    assert_type(noneset_list, NoneSet[Union[str, float]])  # type: ignore[assert-type]  # False-positive in mypy# int and float are merged in generic unions

    assert_type(length_tuple, Length[Tuple[str, str]])
    assert_type(length_list, Length[List[str]])

    assert_type(match_pattern_str_default, MatchPattern[str, Literal[False]])
    assert_type(match_pattern_str, MatchPattern[str, Literal[False]])
    assert_type(match_pattern_str_none, MatchPattern[str, Literal[True]])
    assert_type(match_pattern_bytes_default, MatchPattern[ReadableBuffer, Literal[False]])
    assert_type(match_pattern_bytes, MatchPattern[ReadableBuffer, Literal[False]])
    assert_type(match_pattern_bytes_none, MatchPattern[ReadableBuffer, Literal[True]])

    assert_type(convertible_default, Convertible[int, Literal[False]])
    assert_type(convertible_not_none, Convertible[int, Literal[False]])
    assert_type(convertible_none, Convertible[int, Literal[True]])

    assert_type(minmax_default, MinMax[float, Literal[False]])
    assert_type(minmax_float, MinMax[float, Literal[False]])
    assert_type(minmax_float_none, MinMax[float, Literal[True]])
    assert_type(minmax_int, MinMax[int, Literal[False]])
    assert_type(minmax_int_none, MinMax[int, Literal[True]])

    assert_type(bool_default, Bool[Literal[False]])
    assert_type(bool_not_none, Bool[Literal[False]])
    assert_type(bool_none, Bool[Literal[True]])

    assert_type(datetime_default, DateTime[Literal[False]])
    assert_type(datetime_not_none, DateTime[Literal[False]])
    assert_type(datetime_none, DateTime[Literal[True]])

    assert_type(string_default, String[Literal[False]])
    assert_type(string_not_none, String[Literal[False]])
    assert_type(string_none, String[Literal[True]])

    assert_type(float_default, Float[Literal[False]])
    assert_type(float_not_none, Float[Literal[False]])
    assert_type(float_none, Float[Literal[True]])

    assert_type(integer_default, Integer[Literal[False]])
    assert_type(integer_not_none, Integer[Literal[False]])
    assert_type(integer_none, Integer[Literal[True]])


with_descriptors = WithDescriptors()


# Test with missing subclass
class NotSerialisable:
    descriptor = Descriptor[Any]()


NotSerialisable().descriptor = None  # type: ignore


# Test with Strict subclass
class WithDescriptorsStrict(Strict):
    descriptor = Descriptor[Any]()


WithDescriptorsStrict().descriptor = None


# Test getters
assert_type(with_descriptors.descriptor, str)

assert_type(with_descriptors.typed_not_none, str)
assert_type(with_descriptors.typed_none, Union[str, None])

assert_type(with_descriptors.set_tuple, Union[Literal["a", 1], float])  # type: ignore[assert-type]  # False-positive in mypy
assert_type(with_descriptors.set_list, Union[str, int, float])  # type: ignore[assert-type]  # False-positive in mypy  # Literals are simplified in non-tuples
assert_type(with_descriptors.set_tuple_none, Union[Literal["a", 1, None], float])  # type: ignore[assert-type]  # False-positive in mypy

assert_type(with_descriptors.noneset_tuple, Union[Literal["a", 1], float, None])  # type: ignore[assert-type]  # False-positive in mypy
assert_type(with_descriptors.noneset_list, Union[str, float, None])  # type: ignore[assert-type]  # False-positive in mypy  # int and float are merged in generic unions

assert_type(with_descriptors.length_tuple, Tuple[str, str])
assert_type(with_descriptors.length_list, List[str])

assert_type(with_descriptors.match_pattern_str, str)
assert_type(with_descriptors.match_pattern_str_none, Union[str, None])
assert_type(with_descriptors.match_pattern_bytes, ReadableBuffer)
assert_type(with_descriptors.match_pattern_bytes_none, Union[ReadableBuffer, None])

assert_type(with_descriptors.convertible_not_none, int)
assert_type(with_descriptors.convertible_none, Union[int, None])

assert_type(with_descriptors.minmax_float, float)
assert_type(with_descriptors.minmax_float_none, Union[float, None])
assert_type(with_descriptors.minmax_int, int)
assert_type(with_descriptors.minmax_int_none, Union[int, None])

assert_type(with_descriptors.bool_not_none, bool)
assert_type(with_descriptors.bool_none, Union[bool, None])

assert_type(with_descriptors.datetime_not_none, datetime)
assert_type(with_descriptors.datetime_none, Union[datetime, None])

assert_type(with_descriptors.string_not_none, str)
assert_type(with_descriptors.string_none, Union[str, None])

assert_type(with_descriptors.float_not_none, float)
assert_type(with_descriptors.float_none, Union[float, None])

assert_type(with_descriptors.integer_not_none, int)
assert_type(with_descriptors.integer_none, Union[int, None])


# Test setters (expected type, None, unexpected type)
with_descriptors.descriptor = ""
with_descriptors.descriptor = None  # type: ignore
with_descriptors.descriptor = 0  # type: ignore


with_descriptors.typed_not_none = ""
with_descriptors.typed_not_none = None  # type: ignore
with_descriptors.typed_not_none = 0  # type: ignore

with_descriptors.typed_none = ""
with_descriptors.typed_none = None
with_descriptors.typed_none = 0  # type: ignore


# NOTE: Can't check Set for literal int wen used with a float because any int is a valid float
with_descriptors.set_tuple = "a"
with_descriptors.set_tuple = 0
with_descriptors.set_tuple = 0.0
with_descriptors.set_tuple = None  # pyright: ignore[reportAttributeAccessIssue] # false negative in mypy
with_descriptors.set_tuple = "none"  # pyright: ignore[reportAttributeAccessIssue] # false negative in mypy
with_descriptors.set_tuple = object()  # pyright: ignore[reportAttributeAccessIssue] # false negative in mypy

with_descriptors.set_list = "a"
with_descriptors.set_list = 0
with_descriptors.set_list = 0.0
with_descriptors.set_list = None  # pyright: ignore[reportAttributeAccessIssue] # false negative in mypy
with_descriptors.set_list = "none"  # can't check literals validity
with_descriptors.set_list = object()  # pyright: ignore[reportAttributeAccessIssue] # false negative in mypy

with_descriptors.set_tuple_none = "a"
with_descriptors.set_tuple_none = 0
with_descriptors.set_tuple_none = 0.0
with_descriptors.set_tuple_none = None
with_descriptors.set_tuple_none = "none"  # pyright: ignore[reportAttributeAccessIssue] # false negative in mypy
with_descriptors.set_tuple_none = object()  # type: ignore


with_descriptors.noneset_tuple = "a"
with_descriptors.noneset_tuple = 0
with_descriptors.noneset_tuple = 0.0
with_descriptors.noneset_tuple = None
with_descriptors.noneset_tuple = "none"
with_descriptors.noneset_tuple = object()  # pyright: ignore[reportAttributeAccessIssue] # false negative in mypy

with_descriptors.noneset_list = "a"
with_descriptors.noneset_list = 0
with_descriptors.noneset_list = 0.0
with_descriptors.noneset_list = None
with_descriptors.noneset_list = "none"
with_descriptors.noneset_list = object()  # pyright: ignore[reportAttributeAccessIssue] # false negative in mypy


# NOTE: Can't validate tuple length in a generic manner
with_descriptors.length_tuple = ("a", "a")
with_descriptors.length_tuple = None  # type: ignore
with_descriptors.length_tuple = ["a", "a"]  # type: ignore
with_descriptors.length_tuple = ""  # type: ignore

with_descriptors.length_list = ["a", "a"]
with_descriptors.length_list = None  # type: ignore
with_descriptors.length_list = ("a", "a")  # type: ignore
with_descriptors.length_list = ""  # type: ignore


with_descriptors.match_pattern_str = ""
with_descriptors.match_pattern_str = None  # type: ignore
with_descriptors.match_pattern_str = b""  # type: ignore
with_descriptors.match_pattern_str = 0  # type: ignore

with_descriptors.match_pattern_str_none = ""
with_descriptors.match_pattern_str_none = None
with_descriptors.match_pattern_str_none = b""  # type: ignore
with_descriptors.match_pattern_str_none = 0  # type: ignore

with_descriptors.match_pattern_bytes = b""
with_descriptors.match_pattern_bytes = None  # type: ignore
with_descriptors.match_pattern_bytes = ""  # type: ignore
with_descriptors.match_pattern_bytes = 0  # type: ignore

with_descriptors.match_pattern_bytes_none = b""
with_descriptors.match_pattern_bytes_none = None
with_descriptors.match_pattern_bytes_none = ""  # type: ignore
with_descriptors.match_pattern_bytes_none = 0  # type: ignore


with_descriptors.convertible_not_none = 0
with_descriptors.convertible_not_none = "0"
with_descriptors.convertible_not_none = None  # type: ignore
with_descriptors.convertible_not_none = object()  # type: ignore

with_descriptors.convertible_none = 0
with_descriptors.convertible_none = "0"
with_descriptors.convertible_none = None
with_descriptors.convertible_none = object()  # FIXME: False negative(?) in pyright and mypy


with_descriptors.minmax_float = 0
with_descriptors.minmax_float = "0"
with_descriptors.minmax_float = 0.0
with_descriptors.minmax_float = None  # type: ignore
with_descriptors.minmax_float = object()  # type: ignore

with_descriptors.minmax_float_none = 0
with_descriptors.minmax_float_none = "0"
with_descriptors.minmax_float_none = 0.0
with_descriptors.minmax_float_none = None
with_descriptors.minmax_float_none = object()  # type: ignore

with_descriptors.minmax_int = 0
with_descriptors.minmax_int = "0"
with_descriptors.minmax_int = 0.0
with_descriptors.minmax_int = None  # type: ignore
with_descriptors.minmax_int = object()  # type: ignore

with_descriptors.minmax_int_none = 0
with_descriptors.minmax_int_none = "0"
with_descriptors.minmax_int_none = 0.0
with_descriptors.minmax_int_none = None
with_descriptors.minmax_int_none = object()  # type: ignore


with_descriptors.bool_not_none = False
with_descriptors.bool_not_none = "0"
with_descriptors.bool_not_none = 0
with_descriptors.bool_not_none = None
with_descriptors.bool_not_none = 0.0  # type: ignore
with_descriptors.bool_not_none = object()  # type: ignore

with_descriptors.bool_none = False
with_descriptors.bool_none = "0"
with_descriptors.bool_none = 0
with_descriptors.bool_none = None
with_descriptors.bool_none = 0.0  # type: ignore
with_descriptors.bool_none = object()  # type: ignore


with_descriptors.datetime_not_none = datetime(0, 0, 0)
with_descriptors.datetime_not_none = ""
with_descriptors.datetime_not_none = None  # type: ignore
with_descriptors.datetime_not_none = 0  # type: ignore
with_descriptors.datetime_not_none = date(0, 0, 0)  # type: ignore
with_descriptors.datetime_not_none = time()  # type: ignore

with_descriptors.datetime_none = datetime(0, 0, 0)
with_descriptors.datetime_none = ""
with_descriptors.datetime_none = None
with_descriptors.datetime_none = 0  # type: ignore
with_descriptors.datetime_none = date(0, 0, 0)  # type: ignore
with_descriptors.datetime_none = time()  # type: ignore


with_descriptors.string_not_none = ""
with_descriptors.string_not_none = None  # type: ignore
with_descriptors.string_not_none = 0  # type: ignore

with_descriptors.string_none = ""
with_descriptors.string_none = None
with_descriptors.string_none = 0  # type: ignore


with_descriptors.float_not_none = 0
with_descriptors.float_not_none = 0.0
with_descriptors.float_not_none = "0"
with_descriptors.float_not_none = b"0"
with_descriptors.float_not_none = None  # type: ignore
with_descriptors.float_not_none = object()  # type: ignore

with_descriptors.float_none = 0
with_descriptors.float_none = 0.0
with_descriptors.float_none = "0"
with_descriptors.float_none = b"0"
with_descriptors.float_none = None
with_descriptors.float_none = object()  # FIXME: False negative(?) in pyright and mypy


with_descriptors.integer_not_none = 0
with_descriptors.integer_not_none = 0.0
with_descriptors.integer_not_none = "0"
with_descriptors.integer_not_none = b"0"
with_descriptors.integer_not_none = None  # type: ignore
with_descriptors.integer_not_none = object()  # type: ignore

with_descriptors.integer_none = 0
with_descriptors.integer_none = 0.0
with_descriptors.integer_none = "0"
with_descriptors.integer_none = b"0"
with_descriptors.integer_none = None
with_descriptors.integer_none = object()  # FIXME: False negative(?) in pyright and mypy



from enum import Enum

class PARTSTAT(Enum):
    NEEDS_ACTION = "NEEDS-ACTION"
    ACCEPTED = "ACCEPTED"
    DECLINED = "DECLINED"
    TENTATIVE = "TENTATIVE"
    DELEGATED = "DELEGATED"
    COMPLETED = "COMPLETED"
    IN_PROCESS = "IN-PROCESS"

class FBTYPE(Enum):
    FREE = "FREE"
    BUSY = "BUSY"
    BUSY_UNAVAILABLE = "BUSY-UNAVAILABLE"
    BUSY_TENTATIVE = "BUSY-TENTATIVE"

class CUTYPE(Enum):
    INDIVIDUAL = "INDIVIDUAL"
    GROUP = "GROUP"
    RESOURCE = "RESOURCE"
    ROOM = "ROOM"
    UNKNOWN = "UNKNOWN"

class RELTYPE(Enum):
    PARENT = "PARENT"
    CHILD = "CHILD"
    SIBLING = "SIBLING"

class RANGE(Enum):
    THISANDFUTURE = "THISANDFUTURE"
    THISANDPRIOR = "THISANDPRIOR"  # deprecated

class RELATED(Enum):
    START = "START"
    END = "END"

class ROLE(Enum):
    CHAIR = "CHAIR"
    REQ_PARTICIPANT = "REQ-PARTICIPANT"
    OPT_PARTICIPANT = "OPT-PARTICIPANT"
    NON_PARTICIPANT = "NON-PARTICIPANT"

__all__ = ["PARTSTAT", "FBTYPE", "CUTYPE", "RANGE", "RELATED", "ROLE", "RELTYPE"]



from _typeshed.wsgi import StartResponse, WSGIEnvironment
from collections.abc import Iterable
from http.client import HTTPConnection, HTTPMessage, HTTPSConnection
from typing import ClassVar

__all__ = ["send_request_app", "SendRequest"]

class SendRequest:
    def __init__(self, HTTPConnection: type[HTTPConnection] = ..., HTTPSConnection: type[HTTPSConnection] = ...) -> None: ...
    def __call__(self, environ: WSGIEnvironment, start_response: StartResponse) -> Iterable[bytes]: ...
    filtered_headers: ClassVar[tuple[str, ...]]
    def parse_headers(self, message: HTTPMessage) -> list[tuple[str, str]]: ...

send_request_app: SendRequest



from _typeshed import Incomplete

from aws_xray_sdk.ext.sqlalchemy.query import XRaySession

class XRayBaseQuery: ...

class XRaySignallingSession(XRaySession):
    app: Incomplete
    def __init__(self, db, autocommit: bool = False, autoflush: bool = True, **options) -> None: ...
    def get_bind(self, mapper=None, clause=None): ...

class XRayFlaskSqlAlchemy:
    def __init__(
        self,
        app=None,
        use_native_unicode: bool = True,
        session_options=None,
        metadata=None,
        query_class: type = ...,
        model_class: type = ...,
    ) -> None: ...
    def create_session(self, options: dict[str, Incomplete]): ...



from _typeshed import Incomplete
from collections.abc import Callable, Generator

from networkx.classes.graph import Graph, _Node
from networkx.utils.backends import _dispatchable

__all__ = ["girvan_newman"]

@_dispatchable
def girvan_newman(
    G: Graph[_Node], most_valuable_edge: Callable[..., Incomplete] | None = None
) -> Generator[Incomplete, None, Incomplete]: ...



from _typeshed import Incomplete

from ..rest import RestClientOptions
from ..types import TimeoutType

class Hooks:
    domain: Incomplete
    protocol: Incomplete
    client: Incomplete
    def __init__(
        self,
        domain: str,
        token: str,
        telemetry: bool = True,
        timeout: TimeoutType = 5.0,
        protocol: str = "https",
        rest_options: RestClientOptions | None = None,
    ) -> None: ...
    def all(
        self,
        enabled: bool = True,
        fields: list[str] | None = None,
        include_fields: bool = True,
        page: int | None = None,
        per_page: int | None = None,
        include_totals: bool = False,
    ): ...
    async def all_async(
        self,
        enabled: bool = True,
        fields: list[str] | None = None,
        include_fields: bool = True,
        page: int | None = None,
        per_page: int | None = None,
        include_totals: bool = False,
    ): ...
    def create(self, body: dict[str, Incomplete]) -> dict[str, Incomplete]: ...
    async def create_async(self, body: dict[str, Incomplete]) -> dict[str, Incomplete]: ...
    def get(self, id: str, fields: list[str] | None = None) -> dict[str, Incomplete]: ...
    async def get_async(self, id: str, fields: list[str] | None = None) -> dict[str, Incomplete]: ...
    def delete(self, id: str): ...
    async def delete_async(self, id: str): ...
    def update(self, id: str, body: dict[str, Incomplete]) -> dict[str, Incomplete]: ...
    async def update_async(self, id: str, body: dict[str, Incomplete]) -> dict[str, Incomplete]: ...
    def get_secrets(self, id: str) -> dict[str, Incomplete]: ...
    async def get_secrets_async(self, id: str) -> dict[str, Incomplete]: ...
    def add_secrets(self, id: str, body: dict[str, Incomplete]) -> dict[str, Incomplete]: ...
    async def add_secrets_async(self, id: str, body: dict[str, Incomplete]) -> dict[str, Incomplete]: ...
    def delete_secrets(self, id: str, body: list[str]): ...
    async def delete_secrets_async(self, id: str, body: list[str]): ...
    def update_secrets(self, id: str, body: dict[str, Incomplete]) -> dict[str, Incomplete]: ...
    async def update_secrets_async(self, id: str, body: dict[str, Incomplete]) -> dict[str, Incomplete]: ...



from _typeshed import Incomplete
from collections.abc import Iterable

from networkx.utils.backends import _dispatchable

__all__ = ["join_trees"]

@_dispatchable
def join_trees(rooted_trees: Iterable[Incomplete], *, label_attribute: str | None = None, first_label: int | None = 0): ...



from typing import Final

XK_leftcaret: Final = 0xBA3
XK_rightcaret: Final = 0xBA6
XK_downcaret: Final = 0xBA8
XK_upcaret: Final = 0xBA9
XK_overbar: Final = 0xBC0
XK_downtack: Final = 0xBC2
XK_upshoe: Final = 0xBC3
XK_downstile: Final = 0xBC4
XK_underbar: Final = 0xBC6
XK_jot: Final = 0xBCA
XK_quad: Final = 0xBCC
XK_uptack: Final = 0xBCE
XK_circle: Final = 0xBCF
XK_upstile: Final = 0xBD3
XK_downshoe: Final = 0xBD6
XK_rightshoe: Final = 0xBD8
XK_leftshoe: Final = 0xBDA
XK_lefttack: Final = 0xBDC
XK_righttack: Final = 0xBFC
